@@ -21,7 +21,7 @@ package Spreadsheet::Read;
 use strict;
 use warnings;
 
-our $VERSION = "0.20";
+our $VERSION = "0.24";
 sub  Version { $VERSION }
 
 use Exporter;
@@ -29,21 +29,26 @@ our @ISA       = qw( Exporter );
 our @EXPORT    = qw( ReadData cell2cr cr2cell );
 our @EXPORT_OK = qw( parses rows );
 
-use File::Temp           qw( );
+use File::Temp   qw( );
 use Data::Dumper;
 
-my %can = map { $_ => 0 } qw( csv sxc xls prl );
-for (	[ csv	=> "Text::CSV_XS"		],
-#	[ csv	=> "Text::CSV"			],	# NYI
-#	[ csv	=> "Text::CSV_PP"		],	# NYI
-	[ ods	=> "Spreadsheet::ReadSXC"	],
-	[ sxc	=> "Spreadsheet::ReadSXC"	],
-	[ xls	=> "Spreadsheet::ParseExcel"	],
-	[ prl	=> "Spreadsheet::Perl"		],
-	) {
+my @parsers = (
+    [ csv	=> "Text::CSV_XS"		],
+    [ csv	=> "Text::CSV_PP"		],
+#   [ csv	=> "Text::CSV"			],	# NYI
+    [ ods	=> "Spreadsheet::ReadSXC"	],
+    [ sxc	=> "Spreadsheet::ReadSXC"	],
+    [ xls	=> "Spreadsheet::ParseExcel"	],
+    [ prl	=> "Spreadsheet::Perl"		],
+
+    # Helper modules
+    [ ios	=> "IO::Scalar"			],
+    );
+my %can = map { $_->[0] => 0 } @parsers;
+for (@parsers) {
     my ($flag, $mod) = @$_;
     $can{$flag} and next;
-    eval "require $mod; \$can{$flag} = '$mod'";
+    eval "require $mod; \@_ or \$can{\$flag} = '$mod'";
     }
 $can{sc} = 1;	# SquirelCalc is built-in
 
@@ -202,8 +207,8 @@ sub ReadData ($;@)
 	my $csv;
 	my @data = (
 	    {	type	=> "csv",
-		parser  => "Text::CSV_XS",
-		version	=> $Text::CSV_XS::VERSION,
+		parser  => $can{csv},
+		version	=> $can{csv}->VERSION,
 		quote   => '"',
 		sepchar => ',',
 		sheets	=> 1,
@@ -231,49 +236,71 @@ sub ReadData ($;@)
 					  ","  ;
 	my ($eol) = m{([\r\n]+)\z};
 	$debug > 1 and print STDERR "CSV sep_char '$sep', quote_char '$quo'\n";
-	$csv = Text::CSV_XS->new ({
-	    sep_char   => ($data[0]{sepchar} = $sep),
-	    quote_char => ($data[0]{quote}   = $quo),
-	    eol        => $eol,
-	    get_flags  => 1,
-	    binary     => 1,
-	    });
+	$csv = $can{csv}->new ({
+	    sep_char       => ($data[0]{sepchar} = $sep),
+	    quote_char     => ($data[0]{quote}   = $quo),
+	    eol            => $eol,
+	    keep_meta_info => 1,	# Ignored for Text::CSV_XS <= 0.27
+	    binary         => 1,
+	    }) or die "Cannot create a csv ('$sep', '$quo', '$eol') parser!";
 
+	# while ($row = $csv->getline () {
+	# doesn't work, because I have to fetch the first line for auto
+	# detecting sep and quote
 	$csv->parse ($_);
 	my $row = [ $csv->fields ];
 	do {
-	    my @row = @$row or next;
-	    my $r = ++$data[1]{maxrow};
-	    @row > $data[1]{maxcol} and $data[1]{maxcol} = @row;
-	    foreach my $c (0 .. $#row) {
-		my $val = $row[$c];
-		my $cell = cr2cell ($c + 1, $r);
-		$opt{rc}    and $data[1]{cell}[$c + 1][$r] = $val;
-		$opt{cells} and $data[1]{$cell} = $val;
-		$opt{attr}  and $data[1]{attr}[$c + 1][$r] = { @def_attr };
+	    if (my @row = @$row) {
+		my $r = ++$data[1]{maxrow};
+		@row > $data[1]{maxcol} and $data[1]{maxcol} = @row;
+		foreach my $c (0 .. $#row) {
+		    my $val = $row[$c];
+		    my $cell = cr2cell ($c + 1, $r);
+		    $opt{rc}    and $data[1]{cell}[$c + 1][$r] = $val;
+		    $opt{cells} and $data[1]{$cell} = $val;
+		    $opt{attr}  and $data[1]{attr}[$c + 1][$r] = { @def_attr };
+		    }
 		}
-	    } while ($row = $csv->getline ($in));
+	    else {
+		$csv = undef;
+		}
+	    } while ($csv && ($row = $csv->getline ($in)));
+	close $in;
 
 	for (@{$data[1]{cell}}) {
 	    defined $_ or $_ = [];
 	    }
-	close $in;
 	return _clipsheets $opt{clip}, [ @data ];
 	}
 
     # From /etc/magic: Microsoft Office Document
+    my $xls_from_txt;
     if ($txt =~ m/^(\376\067\0\043
 		   |\320\317\021\340\241\261\032\341
 		   |\333\245-\0\0\0)/x) {
 	$can{xls} or die "Spreadsheet::ParseExcel not installed";
-	$tmpfile = File::Temp->new (SUFFIX => ".xls", UNLINK => 1);
-	print $tmpfile $txt;
-	$txt = "$tmpfile";
+	if ($can{ios}) { # Do not use a temp file if IO::Scalar is available
+	    $xls_from_txt = \$txt;
+	    }
+	else {
+	    $tmpfile = File::Temp->new (SUFFIX => ".xls", UNLINK => 1);
+	    binmode $tmpfile;
+	    print   $tmpfile $txt;
+	    $txt = "$tmpfile";
+	    }
 	}
-    if ($txt =~ m/\.xls$/i and -f $txt) {
+    if ($xls_from_txt or $txt =~ m/\.xls$/i && -f $txt) {
 	$can{xls} or die "Spreadsheet::ParseExcel not installed";
-	$debug and print STDERR "Opening XLS $txt\n";
-	my $oBook = Spreadsheet::ParseExcel::Workbook->Parse ($txt);
+	my $oBook;
+	if ($xls_from_txt) {
+	    $debug and print STDERR "Opening XLS \$txt\n";
+	    $oBook = Spreadsheet::ParseExcel::Workbook->Parse (\$txt);
+	    }
+	else {
+	    $debug and print STDERR "Opening XLS $txt\n";
+	    $oBook = Spreadsheet::ParseExcel::Workbook->Parse ($txt);
+	    }
+	$oBook or return;
 	$debug > 8 and print STDERR Data::Dumper->Dump ([$oBook],["oBook"]);
 	my @data = ( {
 	    type	=> "xls",
@@ -476,7 +503,7 @@ sub ReadData ($;@)
 
 =head1 DESCRIPTION
 
-Spreadsheet::Read tries to transparantly read *any* spreadsheet and
+Spreadsheet::Read tries to transparently read *any* spreadsheet and
 return its content in a universal manner independent of the parsing
 module that does the actual spreadsheet scanning.
 
@@ -484,8 +511,8 @@ For OpenOffice this module uses Spreadsheet::ReadSXC
 
 For Excel this module uses Spreadsheet::ParseExcel
 
-For CSV this module uses Text::CSV_XS. Support for Text::CSV_PP is
-planned.
+For CSV this module uses Text::CSV_XS (0.29 or up prefered) or
+Text_PP (1.05 or up required).
 
 For SquirrelCalc there is a very simplistic built-in parser
 
@@ -529,7 +556,7 @@ The C<cell> hash entry contains unformatted data, while the hash entries
 with the traditional labels contain the formatted values (if applicable).
 
 The control hash (the first entry in the returned array ref), contains
-some spreadsheet metadata. The entry C<sheet> is there to be able to find
+some spreadsheet meta-data. The entry C<sheet> is there to be able to find
 the sheets when accessing them by name:
 
   my %sheet2 = %{$ref->[$ref->[0]{sheet}{"Sheet 2"}]};
@@ -555,8 +582,9 @@ the sheets when accessing them by name:
 Tries to convert the given file, string, or stream to the data
 structure described above.
 
-Precessing data from a stream or content is supported for Excel (through a
-File::Temp temporary file), or for XML (OpenOffice), but not for CSV.
+Precessing data from a stream or content is supported for Excel
+(through a File::Temp temporary file or IO::Scalar when available),
+or for XML (OpenOffice), but not for CSV.
 
 ReadSXC does preserve sheet order as of version 0.20.
 
@@ -737,12 +765,6 @@ names C<meta>, or just be new values in the C<attr> hashes.
 
 I consider adding any spreadsheet interface that offers a usable API.
 
-=item Safety / flexibility
-
-Now that the different parsers are only activated if the module can be
-loaded, we need more flexibility is switching from Text::CSV_XS to
-Text::CSV_PP in the parser part.
-
 =item OO-ify
 
 Consider making the ref an object, though I currently don't see the big
@@ -755,7 +777,7 @@ OO interface.
 
 =over 2
 
-=item Text::CSV_XS
+=item Text::CSV_XS, Text::CSV_PP
 
 http://search.cpan.org/~hmbrand/
 
